#!/usr/bin/env python3
import logging
from threading import Thread, Event

class Transport:
    """
    PyTesira transport base class

    This acts as a starting point for transport-type-specific channel implementations (e.g., SSH or Serial)
    which has their own methods and requirements. Note that authentication schemes are implemented directly
    in the subclass (since this can be very different from method to method)
    """

    def __init__(self):
        """
        Initialize base transport class
        """

        # Read buffer size (bytes)
        self.read_buffer_size = 4096

        # "Connected" text that confirms transport is ready
        self.ttp_welcome = "Welcome to the Tesira Text Protocol Server..." 

    def start(self,
        exit_event: Event,         # Core exit event (terminates transport channel too)
        connected_flag: Event,     # Connected flag, lets us tell everyone else the connection status
    ):
        """
        Actually start the backend transport. This will be called by the DSP class and is not
        intended to be called directly elsewhere
        """
        raise NotImplementedError

    @property
    def recv_ready(self) -> bool:
        """
        Data ready in read buffer?
        """
        raise NotImplementedError

    def recv(self, buffer_size : int) -> str:
        """
        Read data from RX buffer
        """
        raise NotImplementedError

    def send(self, data : str) -> None:
        """
        Send data to device
        """
        raise NotImplementedError


#!/usr/bin/env python3
from pytesira.transport.transport import Transport
from threading import Thread, Event
import paramiko
import logging
import traceback
import time

class SSH(Transport):
    """
    PyTesira SSH transport using Paramiko
    """

    def __init__(self,
        hostname: str,                      # Device hostname/IP
        username: str,                      # SSH username
        password: str,                      # SSH password
        port: int = 22,                     # SSH port
        host_key_check: bool = True,        # Enable SSH host key checking?
        connection_timeout: int = 5,        # SSH connection timeout (seconds)
    ) -> None:

        # Base class initialization
        super().__init__()

        # Logger
        self.__logger = logging.getLogger(__name__)

        # Save connection parameters
        self.__hostname = str(hostname)
        self.__username = str(username)
        self.__password = str(password)
        self.__port = int(port)
        self.__host_key_check = bool(host_key_check)
        self.__connection_timeout = int(connection_timeout)
        assert 1 <= self.__port <= 65535, f"Invalid SSH port {self.__port}"
        assert 1 <= self.__connection_timeout, f"Invalid connection timeout {self.__connection_timeout}"

        # Actual connection object (Paramiko channel)
        self.__channel = None

        # Stop Paramiko from flooding the terminal with its debug messages
        # (we typically don't have to debug that deep...)
        logging.getLogger("paramiko.transport").setLevel(logging.INFO)

    # =================================================================================================================

    def start(self,
        exit_event: Event,                  # Core exit event (terminates transport channel too)
        connected_flag: Event,              # Connected flag, lets us tell everyone else the connection status
    ) -> None:
        """
        Actually start the backend transport
        """

        # Save handles to those important event flags!
        self.__exit = exit_event
        self.__connected = connected_flag

        # Start transport thread
        self.__thread = Thread(target = self.__mainThread)
        self.__thread.start()

    # =================================================================================================================

    def assert_device_ready(f):
        """
        Decorator for calls that requires the device to be ready
        """

        def wrapper(self, *args, **kwargs):
            assert self.__channel != None, "Channel not initialized"
            assert self.__channel.active, "Device not ready"
            return f(self, *args, **kwargs)

        return wrapper

    # =================================================================================================================

    @property
    @assert_device_ready
    def recv_ready(self) -> bool:
        """
        Data ready in read buffer?
        """
        return self.__channel.recv_ready()

    # =================================================================================================================

    @assert_device_ready
    def recv(self, buffer_size : int) -> str:
        """
        Read data from RX buffer
        """
        return str(self.__channel.recv(buffer_size).decode())

    # =================================================================================================================

    @assert_device_ready
    def send(self, data : str) -> None:
        """
        Send data to device
        """
        self.__channel.send(f"{data}\n")

    # =================================================================================================================

    def __mainThread(self) -> None:
        """
        Main thread (runs forever until we're told to exit)
        """
        while not self.__exit.is_set():

            try:

                # If channel isn't connected, we connect
                if (not self.__channel) or (self.__channel is not None and self.__channel.closed):
                    self.__connect()

                # Now the thread doesn't have to do anything, since Paramiko will take over
                # we just sleep and stay out of the way
                time.sleep(0.5)

            except Exception as e:
                # Oh no, something bad happened. We log that, and wait for a few seconds
                # before trying to connect again
                self.__connected.clear()
                self.__logger.error(f"Thread exception: {e} ({traceback.format_exc()})")
                time.sleep(2)

        # If we're here, it means the global exit flag has been set,
        # and it's our cue to terminate, so we first close the backend channel if possible:
        if self.__channel is not None:
            self.__channel.close()

        # Then we exit
        return

    # =================================================================================================================

    def __connect(self) -> None:
        """
        Connect (or re-connect) to the SSH channel
        """

        # If there's a lingering session, we'll try to close that if possible. If not - not a big deal:
        try:
            self.__channel.close()
            self.__logger.debug(f"old SSH channel closed")
        except:
            pass

        # Start from disconnected state
        self.__connected.clear()

        # Session and channel (Paramiko shell session)
        self.__session = paramiko.SSHClient()    
        self.__session.set_missing_host_key_policy(paramiko.RejectPolicy() if self.__host_key_check else paramiko.WarningPolicy())
        self.__session.connect(
            self.__hostname, 
            self.__port, 
            username = self.__username, 
            password = self.__password, 
            timeout = self.__connection_timeout
        )
        self.__channel = self.__session.invoke_shell()
        self.__logger.debug(f"channel started")

        # Try to connect and wait until we either get the welcome text, or reached timeout
        # limitations, whichever one comes first
        conn_init_time = time.perf_counter()
        welcomed = False
        self.__logger.info(f"waiting for session establishment")
        while time.perf_counter() - conn_init_time < self.__connection_timeout:
            if self.__channel.active:
                time.sleep(0.01)
                if self.__channel.recv_ready():
                    received = self.__channel.recv(self.read_buffer_size).decode()
                    if self.ttp_welcome in received:
                        welcomed = True
                        break
        
        # If there's anything left in the buffer, clear it out (for good measure)
        while self.__channel.recv_ready():
            _ = self.__channel.recv(self.read_buffer_size).decode()

        if not welcomed:
            # Uh oh, we didn't get a valid response from the DSP
            raise Exception(f"timeout waiting for session establishment")
            self.__connected.clear()
        else:
            # Connection OK :)
            self.__logger.info(f"Tesira text protocol session established ({round(time.perf_counter() - conn_init_time, 3)} sec)")
            self.__connected.set()


#!/usr/bin/env python3
from threading import Event
from pytesira.block.block import Block
from queue import Queue
from pytesira.util.ttp_response import TTPResponse, TTPResponseType
import time
import logging

class MuteControl(Block):

    # Define version of this block's code here. A mismatch between this
    # and the value saved in the cached attribute-list value file will
    # trigger a re-discovery of attributes, to handle any changes
    VERSION = "0.1.0"

    # =================================================================================================================

    def __init__(self,
        block_id: str,                  # block ID on Tesira
        exit_flag: Event,               # exit flag to stop the block's threads (sync'd with everything else)                    
        connected_flag: Event,          # connected flag (module can refuse to allow access if this is not set)
        command_queue: Queue,           # command queue (to run synchronous commands and get results)
        subscriptions: dict,            # subscription container on main thread
        init_helper: str|None = None,   # initialization helper (if not specified, query everything from scratch)
    ) -> None:

        # Setup logger
        self._logger = logging.getLogger(f"{__name__}.{block_id}")

        # Initialize base class
        super().__init__(block_id, exit_flag, connected_flag, command_queue, subscriptions, init_helper)

        # If init helper isn't set, this is the time to query
        try:
            assert init_helper is not None, "no helper present"

            self.ganged = init_helper["ganged"]
            self.channels = {}
            for i, d in init_helper["channels"].items():
                self.channels[str(i)] = d

        except Exception as e:
            # There's a problem, throw warning and then simply query
            self._logger.warning(f"cannot use initialization helper: {e}")
            self.__query_attributes()

        # Setup subscriptions
        self.register_subscriptions()
        
    # =================================================================================================================

    def register_subscriptions(self) -> None:
        """
        (re)register subscriptions for this module. This should be called by each
        module on init, and may be called again by the main thread if an interruption
        in connectivity is detected (e.g., SSH disconnect-then-reconnect)
        """
        self._register_subscription(subscribe_type = "mutes", channel = None)

    # =================================================================================================================

    def export_init_helper(self) -> dict:
        """
        Export initialization helper dict (to make setup faster in the future)
        """
        helper = super().export_init_helper()
        helper["helper"] = {
            "ganged" : self.ganged,
            "channels" : self.channels
        }

        return helper

    # =================================================================================================================

    def subscription_callback(self, response : TTPResponse) -> None:
        """
        Handle incoming subscription callbacks
        """

        # Mutes?
        if response.subscription_type == "mutes":
            for i, mute in enumerate(response.value):
                idx = i + 1
                if str(idx) in self.channels.keys():
                    self.channels[str(idx)]["muted"] = bool(mute)
                else:
                    self._logger.error(f"mute response invalid index: {idx}")
            self._logger.debug(f"mute state changed: {response.value}")

        # Huh, this isn't handled?
        else:
            self._logger.debug(f"unhandled subscription callback: {response}")

        # Call superclass handler to deal with callbacks
        super().subscription_callback(response)

    # =================================================================================================================

    def __query_attributes(self) -> None:

        # Ganged setup?
        self.ganged = bool(self._sync_command(f"{self._block_id} get ganged").value)

        # How many channels?
        num_channels = int(self._sync_command(f"{self._block_id} get numChannels").value)
        self.channels = {}

        # For each channel, what's the index and labels?
        # NOTE: Tesira indices starts at 1, in some cases 0 is a special ID meaning all channels
        for i in range(1, num_channels + 1):
            self.channels[str(i)] = {
                "index" : i,
                "label" : self._sync_command(f"{self._block_id} get label {i}").value
            }

    # =================================================================================================================

    def set_mute(self, value : bool, channel : int = 0) -> TTPResponse:
        """
        Stub for set mute, for blocks that supports it
        """
        assert type(value) == bool, "invalid value type for set_mute"
        return self._sync_command(f'"{self._block_id}" set mute {channel} {str(value).lower()}')

    # =================================================================================================================


#!/usr/bin/env python3
from threading import Event
from queue import Queue
from pytesira.util.ttp_response import TTPResponse, TTPResponseType
from collections.abc import Callable
import time

class Block:

    # Define version of this block's code here. A mismatch between this
    # and the value saved in the cached attribute-list value file will
    # trigger a re-discovery of attributes, to handle any changes
    VERSION = "base-0.1.0"

    def __init__(self,
        block_id: str,                  # block ID on Tesira
        exit_flag: Event,               # exit flag to stop the block's threads (sync'd with everything else)                    
        connected_flag: Event,          # connected flag (module can refuse to allow access if this is not set)
        command_queue: Queue,           # command queue (to run synchronous commands and get results)
        subscriptions: dict,            # subscription container on main thread
        init_helper: str|None = None,   # initialization helper (if not specified, query everything from scratch)
    ) -> None:

        # Logger should be set up first by subclass, but if not
        # we can also set it up again here
        if not self._logger:
            self._logger = logging.getLogger(f"{__name__}.{block_id}")
        
        # Load setup variables
        self._block_id = str(block_id)
        self._exit = exit_flag
        self._connected = connected_flag
        self._command_queue = command_queue
        self._init_helper = init_helper
        self._subscriptions = subscriptions

        # Synchronous command mailbox
        self._sync_cmd_mailbox = None

        # Callbacks
        self._callbacks = {}

    def register_subscriptions(self) -> None:
        """
        (re)register subscriptions for this module. This should be called by each
        module on init, and may be called again by the main thread if an interruption
        in connectivity is detected (e.g., SSH disconnect-then-reconnect)
        """
        return

    def subscription_callback(self, response : TTPResponse) -> None:
        """
        Subscription results matching this block ID will be returned here so that
        they can be processed
        """

        # Call handler that should call OUR callbacks as well
        self._on_subscription_callback()

        return

    def subscribe(self) -> None:
        """
        Set up required subscriptions. Can be omitted (base class stub used)
        if the block type does not require any subscriptions
        """
        return

    def export_init_helper(self) -> dict:
        """
        Export initialization helper. This will be called by the main thread
        after the attribute query process, which allows for the block map
        to be updated
        """
        return {
            "version" : self.VERSION,
            "helper" : {}
        }

    def set_mute(self, value : bool, channel : int = 0) -> TTPResponse:
        """
        Stub for set mute, for blocks that supports it
        """
        raise NotImplementedError

    def set_level(self, value : float, channel : int = 0) -> TTPResponse:
        """
        Stub for set audio level, for blocks that supports it
        """
        raise NotImplementedError

    # =================================================================================================================
    # Generally, there's no need to change or override methods below here, they apply to every block in the same way
    # =================================================================================================================

    def register_callback(self, callback : Callable, key : str|None = None) -> None:
        """
        Register a callback, optionally specifying a callback key so it can be 
        easily unregistered if so needed
        """
        self._callbacks[key if key else callback] = callback
        self._logger.debug(f"callback registered ({len(self._callbacks)} active callbacks)")

    def unregister_callback(self, key : str) -> None:
        """
        Unregister callback based on key specified earlier on callback creation
        """
        try:
            del self._callbacks[key]
            self._logger.debug(f"callback removed ({len(self._callbacks)} active callbacks)")
        except Exception as e:
            self._logger.warning(f"cannot remove callback {key}: {e}")

    # =================================================================================================================
    # More internal functions which should only be called by this class and its subclasses
    # again, they should apply to every block in the same way, so no need to edit or override them
    # =================================================================================================================

    def _on_subscription_callback(self):
        """
        Handler for "on subscription callback": checks all locally registered callbacks
        and call them if possible (with current block as the object variable)
        """
        for key, callback in self._callbacks.items():
            if callable(callback):
                callback(self)
                self._logger.debug(f"callback invoked: {callback}")
            else:
                self._logger.debug(f"uncallable callback: {callback} ({type(callback)})")

    def _register_subscription(self, subscribe_type : str, channel : int|None = None) -> None:
        """
        Register subscription with the DSP. This function generates the subscription command
        with the correct prefix IDs and metadata, such that responses will be directed back here

        DO NOT CHANGE this without also double checking what's done in the main thread
        AS WELL AS TTPResponse, otherwise subscribed data might end up in the wrong place!
        """

        # Subscriptions can be done on a per channel basis too, so we need to handle that
        sub_channel = "" if channel is None else f" {channel}"
        channel_id = "ALL" if sub_channel == "" else str(sub_channel).strip()

        # Create subscription name
        sub_name = f"S_{subscribe_type}_{channel_id}_{self._block_id}"

        # Create subscription string
        sub_string = f'"{self._block_id}" subscribe {subscribe_type}{sub_channel} "{sub_name}"'

        # Add that subscription to the main subscription list
        subscription_meta = (self, self._block_id, sub_name, sub_string)
        self._subscriptions[sub_name] = subscription_meta

        # Send command to device to actually start subscription
        self._sync_command(sub_string)
        self._logger.debug(f"subscription setup: {self._block_id} {subscribe_type} (total {len(self._subscriptions)} subscriptions active)")

    def _sync_command_callback(self, data : TTPResponse) -> None:
        """
        Should we queue a synchronous command and that gets executed by the main thread,
        this method gets called for a response to that command
        """
        self._sync_cmd_mailbox = data
        return

    def _sync_command(self, command : str, timeout : float = 3.0) -> TTPResponse:
        """
        Execute synchronous command (technically this abstracts it behind a queue,
        but we don't have to see that elsewhere in the block module code, which makes it nicer!)
        """
        # Reset mailbox, just in case
        self._sync_cmd_mailbox = None

        # Queue command
        self._command_queue.put((self, command))
        cmd_queued = time.perf_counter()

        # Now we wait until we either get data or timeout occurs, whichever comes first
        while time.perf_counter() - cmd_queued < timeout:

            # Let other tasks run
            time.sleep(0.01)

            # Did we get something?
            if self._sync_cmd_mailbox:
                cmd_response = self._sync_cmd_mailbox
                self._sync_cmd_mailbox = None
                return cmd_response

        # If we're here, we got... nothing
        self._sync_cmd_mailbox = None
        raise Exception(f"Command timeout: {command}")


#!/usr/bin/env python3
from threading import Event
from pytesira.block.block import Block
from queue import Queue
from pytesira.util.ttp_response import TTPResponse, TTPResponseType
import time
import logging

class SourceSelector(Block):

    # Define version of this block's code here. A mismatch between this
    # and the value saved in the cached attribute-list value file will
    # trigger a re-discovery of attributes, to handle any changes
    VERSION = "0.1.0"

    # =================================================================================================================

    def __init__(self,
        block_id: str,                  # block ID on Tesira
        exit_flag: Event,               # exit flag to stop the block's threads (sync'd with everything else)                    
        connected_flag: Event,          # connected flag (module can refuse to allow access if this is not set)
        command_queue: Queue,           # command queue (to run synchronous commands and get results)
        subscriptions: dict,            # subscription container on main thread
        init_helper: str|None = None,   # initialization helper (if not specified, query everything from scratch)
    ) -> None:

        # Setup logger
        self._logger = logging.getLogger(f"{__name__}.{block_id}")

        # Initialize base class
        super().__init__(block_id, exit_flag, connected_flag, command_queue, subscriptions, init_helper)

        # If init helper isn't set, this is the time to query
        try:
            assert init_helper is not None, "no helper present"
            self.__load_init_helper(init_helper)

        except Exception as e:
            # There's a problem, throw warning and then simply query
            self._logger.warning(f"cannot use initialization helper: {e}")
            self.__query_attributes()

        # Setup subscriptions
        self.register_subscriptions()
        
    # =================================================================================================================

    def register_subscriptions(self) -> None:
        """
        (re)register subscriptions for this module. This should be called by each
        module on init, and may be called again by the main thread if an interruption
        in connectivity is detected (e.g., SSH disconnect-then-reconnect)
        """
        self._register_subscription(subscribe_type = "outputMute", channel = None)
        self._register_subscription(subscribe_type = "outputLevel", channel = None)
        self._register_subscription(subscribe_type = "sourceSelection", channel = None)

        # Subscribe to source levels too
        for index in self.sources.keys():
            self._register_subscription(subscribe_type = "sourceLevel", channel = index)

    # =================================================================================================================

    def __load_init_helper(self, init_helper : dict) -> None:
        """
        Use initialization helper to set up attributes instead of querying
        """
        self.stereo = bool(init_helper["stereo"])
        self.num_input = int(init_helper["num_input"])
        self.num_output = int(init_helper["num_output"])
        self.muted = False          # updated by subscription, not in helper
        self.selected_source = 0    # updated by subscription, not in helper
        self.sources = init_helper["sources"]
        self.output_level = init_helper["output_level"]

    def export_init_helper(self) -> dict:
        """
        Export initialization helper dict (to make setup faster in the future)
        """
        helper = super().export_init_helper()
        helper["helper"] = {
            "stereo" : self.stereo,
            "num_input" : self.num_input,
            "num_output" : self.num_output,
            "sources" : self.sources,
            "output_level" : self.output_level,
        }

        return helper

    # =================================================================================================================

    def __query_attributes(self) -> None:

        # Stereo mode?
        self.stereo = bool(self._sync_command(f"{self._block_id} get stereoEnable").value)

        # Number of inputs and outputs
        self.num_input = int(self._sync_command(f"{self._block_id} get numInputs").value)
        self.num_output = int(self._sync_command(f"{self._block_id} get numOutputs").value)

        # Output muted?
        self.muted = bool(self._sync_command(f"{self._block_id} get outputMute").value)

        # How many actual input channels (sources) and outputs?
        if self.stereo:
            self.num_input = int(self.num_input // 2)
            self.num_output = int(self.num_output // 2)

        # Which channel is selected? (0 = nothing)
        self.selected_source = 0

        # For each source, they have index, label, and level attributes assigned
        # as well as a "selected" helper attribute to verify which source is currently selected
        # NOTE: Tesira indices starts at 1
        self.sources = {}
        for i in range(1, self.num_input + 1):
            self.sources[str(i)] = {
                "index" : i,
                "label" : self._sync_command(f"{self._block_id} get label {i}").value,
                "level" : {
                    "min" : self._sync_command(f"{self._block_id} get sourceMinLevel {i}").value,
                    "max" : self._sync_command(f"{self._block_id} get sourceMaxLevel {i}").value,
                },
                "selected" : False
            }

        # We also allow control of output levels
        self.output_level = {
            "min" : self._sync_command(f"{self._block_id} get outputMinLevel").value,
            "max" : self._sync_command(f"{self._block_id} get outputMaxLevel").value
        }

    # =================================================================================================================

    def subscription_callback(self, response : TTPResponse) -> None:
        """
        Handle incoming subscription callbacks
        """

        """
        # Subscribe to source levels too
        for index in self.sources.keys():
            self._register_subscription(subscribe_type = "sourceLevel", channel = index)
        """

        # Output mute?
        if response.subscription_type == "outputMute":
            self.muted = bool(response.value)
            self._logger.debug(f"mute state = {response.value}")

        # Output level?
        elif response.subscription_type == "outputLevel":
            self.output_level["current"] = float(response.value)
            self._logger.debug(f"output level = {response.value}")

        # Source selection?
        elif response.subscription_type == "sourceSelection":
            self.selected_source = int(response.value)

            # Update each sources too for easy reference
            for i in self.sources.keys():
                self.sources[i]["selected"] = bool(str(i) == str(self.selected_source))

            self._logger.debug(f"source selection = {response.value}")

        # Source levels?
        elif response.subscription_type == "sourceLevel":
            if str(response.subscription_channel_id) in self.sources.keys():
                self.sources[str(response.subscription_channel_id)]["level"]["current"] = float(response.value)
                self._logger.debug(f"source level update on {response.subscription_channel_id} = {response.value}")
            else:
                self._logger.error(f"source level invalid index: {idx}")

        # Huh, this isn't handled?
        else:
            self._logger.debug(f"unhandled subscription callback: {response}")

        # Call superclass handler to deal with callbacks
        super().subscription_callback(response)

    # =================================================================================================================

    def set_mute(self, value : bool, channel : int = 0) -> TTPResponse:
        """
        Stub for set mute, for blocks that supports it

        Note that we keep the channel parameter to preserve compatibility with
        other set_mute's, but it's not used nor required in this case
        """
        assert type(value) == bool, "invalid value type for set_mute"
        return self._sync_command(f'"{self._block_id}" set outputMute {str(value).lower()}')

    # =================================================================================================================

    def select_source(self, source : int = 0) -> TTPResponse:
        """
        Select a specific source (or specify source = 0 to not select anything)
        """
        assert type(source) == int, "invalid value type for source"
        assert 0 <= source, "invalid value for source"
        return self._sync_command(f'"{self._block_id}" set sourceSelection {source}')

    # =================================================================================================================

    def set_source_level(self, source : int, value : float) -> TTPResponse:
        """
        Set level for a specific source
        """
        assert type(source) == int, "invalid value type for source"
        assert 1 <= source, "invalid value for source"
        assert type(value) == float, "invalid value type for level"
        return self._sync_command(f'"{self._block_id}" set sourceLevel {source} {value}')

    # =================================================================================================================

    def set_level(self, value : float, channel : int = 0) -> TTPResponse:
        """
        Set output level. Note that we keep the calling signature of set_level,
        but ignore channels here
        """
        assert type(value) == float, "invalid type for value"
        return self._sync_command(f'"{self._block_id}" set outputLevel {value}')


#!/usr/bin/env python3
import re
import json
import logging
from enum import Enum

class TTPResponseType(Enum):
    """
    TTP response types
    """
    UNKNOWN = 0             # Unknown / default, should never really occur
    CMD_OK = 1              # OK response with no value (e.g., command acknowledgements)
    CMD_OK_VALUE = 2        # OK response with attached value
    CMD_ERROR = 3           # Error response
    SUBSCRIPTION = 5        # Subscription response

class TTPResponse():
    """
    Tesira Text Protocol (TTP) response object

    Handles parsing of the response text with overly complicated regexes... that's pretty much it

    NOTE: this assumes that the TTP channel is set to verbose mode!
    """

    def __init__(self, ttp_string : str) -> None:
        """
        One stop function that does everything this class is supposed to do

        Give it a raw TTP string, and it'll create a response object with all the
        attributes filled in - easy as that!
        """

        # Logger
        self.__logger = logging.getLogger(f"{__name__}")
        
        # Input string cleaning (and sanity checks)
        ttp_string = str(ttp_string).strip()
        assert ttp_string != "", "empty string given"

        # Save raw string too just in case
        self.raw = ttp_string

        # What's my type (and optionally value?)
        self.type = TTPResponseType.UNKNOWN
        self.value = None
        self.publish_token = None

        # OK response
        if ttp_string.startswith("+OK"):
            # Do I have anything in the response?
            resp = str(ttp_string[3:]).strip()
            if resp == "":
                # Nothing, so we're done here as there's nothing else to process
                self.type = TTPResponseType.CMD_OK
                return
            else:
                # Yes, there's value to be processed
                self.type = TTPResponseType.CMD_OK_VALUE

                # We expect the response to have the format of HEADER_TYPE : VALUE
                # (sometimes the header type is indicative of what the data would look like,
                #  but sometimes it doesn't, so we handle those here)
                header_type = resp.split(":", 1)[0].replace('"', '').strip().lower()
                data = resp.split(":", 1)[1].strip()

                if header_type == "value":

                    # If data is enclosed in double quotes, we strip that
                    if data.startswith('"') and data.endswith('"'):
                        data = data[1:-1]
                    
                    # Parse data
                    self.value = self.__deep_parse_value(data)

                elif header_type == "list":
                    # Extract list items and save (after value conversion of each item) to an attribute
                    items = list(re.findall('"([^"]*)"', data.split("[", 1)[1].split("]", 1)[0].strip()))
                    self.value = [self.__value_format(i) for i in items]

        # Error response
        elif ttp_string.startswith("-ERR"):
            self.type = TTPResponseType.CMD_ERROR
            self.value = str(ttp_string[4:]).strip()

        # Subscription response
        elif ttp_string.startswith("!"):

            self.type = TTPResponseType.SUBSCRIPTION

            # Subscription returns are given as a key-value pair, AS PLAIN TEXT, so we need to decode that
            kv_return = self.__deep_parse_value(str(ttp_string[1:]), force_first_layer_as_dict = True)

            # Now, in kv_return, we have a key-value container. This should have publishToken, which will
            # correspond to the subscription string used to make the subscription, as well as value
            assert "value" in kv_return, f"subscription callback with no data value: {ttp_string}"
            assert "publishToken" in kv_return, f"subscription callback with no publish token: {ttp_string}"

            # Value and publish token data is returned verbatim
            self.value = kv_return["value"]
            self.publish_token = kv_return["publishToken"]

            # We however want to process the publish token a little more too, as it contains valuable
            # data on what it is (format in block.py, should be f"S_{subscribe_type}_{channel_id}_{self._block_id}")
            pub_token_splitted = self.publish_token.split("_", 3)
            assert str(pub_token_splitted[0]).strip() == "S", f"non-prefixed subscription callback: {ttp_string}"

            self.subscription_type = str(pub_token_splitted[1]).strip()
            self.subscription_channel_id = str(pub_token_splitted[2]).strip()
            self.subscription_block_id = str(pub_token_splitted[3]).strip()

        # Unexpected input
        else:
            raise ValueError(f"invalid TTP response type for input: {ttp_string}")
    
    def __repr__(self) -> str:
        """
        String representation of TTP responses
        """
        if self.value:
            if self.publish_token and self.type == TTPResponseType.SUBSCRIPTION:
                return f"Subscription response [{type(self.value)}] {self.value} (type={self.subscription_type}, channel={self.subscription_channel_id}, block={self.subscription_block_id})"
            else:
                return f"Response [{self.type}][{type(self.value)}] {self.value}"
        else:
            return f"Response [{self.type}]"

    def __deep_parse_value(self, raw : str, force_first_layer_as_dict : bool = False) -> dict:
        """
        Given a long string (some sort of proto-JSON which doesn't include commas), we want to parse it and return nice
        objects...
        """
        # Clean up raw values
        raw = str(raw).strip()

        # Now, what type do we have here?
        if (raw.startswith("{") and raw.endswith("}")) or force_first_layer_as_dict:
            # Dict
            raw = str(raw[1:-1]) if not force_first_layer_as_dict else raw

            # Find key-value items in here
            key_values = list(re.findall(r'("[^"]*"|[^: ]+):\s*("[^"]*"|\[.*?\]|\{.*?\}|[^"\[\{ ]*)', raw))

            # Clean up keys and values (strip outermost quotes if one exists)
            cleaned = {}
            for o in key_values:
                key, val = o
                key = str(key).strip()
                val = str(val).strip()

                if key.startswith('"') and key.endswith('"'):
                    key = str(key[1:-1])

                if val.startswith('"') and val.endswith('"'):
                    val = str(val[1:-1]) 

                # Recursively resolve value items
                cleaned[key] = self.__deep_parse_value(val)

            return cleaned

        elif raw.startswith("[") and raw.endswith("]"):
            # List
            raw = str(raw[1:-1])

            # Use a stack to parse list for tokens
            tokens = []
            current_token = ""
            stack = []

            for char in raw + " ":
                if char in "[\"{":
                    stack.append(char)
                    current_token += char
                elif char in "]}\"":
                    if stack and ((char == "]" and stack[-1] == "[") or (char == "}" and stack[-1] == "{") or (char == "\"" and stack[-1] == "\"")):
                        stack.pop()
                        current_token += char
                    else:
                        current_token += char
                elif char.isspace() and not stack:
                    if current_token:
                        tokens.append(current_token)
                        current_token = ""
                else:
                    current_token += char
            
            # Add the last token to the list if it's not empty
            if current_token.strip():
                tokens.append(current_token.strip())
            
            # Recursively process each token
            return [self.__deep_parse_value(t) for t in tokens]

        else:
            # Value (base case)
            if raw.startswith('"') and raw.endswith('"'):
                raw = str(val[1:-1]) 
            return self.__value_format(raw)

    def __value_format(self, val : float|int|bool|str) -> float|int|bool|str:
        """
        Given a data as a value, try to guess whether it's a string, boolean,
        integer, or float... then cast accordingly
        """
        val = str(val).strip()
        try:
            # Attempt conversion to float
            _ = float(val)

            # If we're here without an error thrown already, we have
            # either a float or bool. We check for existence of a decimal
            # point in the original text to figure out what type:
            if "." in val:
                return float(val)
            else:
                return int(val)
        except ValueError as e:
            # Can't be converted to floating point number, so it's probably
            # a boolean or string. We do a simple check here:
            if str(val).lower() in ["true", "yes", "on"]:
                return bool(True)
            elif str(val).lower() in ["false", "no", "off"]:
                return bool(False)
            else:
                return str(val)


#!/usr/bin/env python3
# Tests for the Tesira Text Protocol (TTP) parser
from pytesira.util.ttp_response import TTPResponse, TTPResponseType
import pprint

def test_parse_ok():
    """
    Test parsing of a basic OK message with no response
    """
    r = TTPResponse("+OK")
    assert r.type == TTPResponseType.CMD_OK

def test_parse_ok_long_spaces():
    """
    Test parsing of a basic OK message with a string that contains longer trailing spaces
    (this should never be the case, but we check for that anyway)
    """
    r = TTPResponse("+OK                                           ")
    assert r.type == TTPResponseType.CMD_OK

def test_parse_ok_simple_value():
    """
    Test parsing of a OK response message where there's also a value, but those values
    are relatively simple - singular values of different types
    """
    vals = [False, 1, -1, 2.0, -2.0, "test string"]
    for v in vals:
        v_str = str(v).lower().strip()
        for vq in [f"{v_str}", f'"{v_str}"']:
            r = TTPResponse(f'+OK "value":{vq}')
            assert r.type == TTPResponseType.CMD_OK_VALUE
            assert r.value == v
            assert type(r.value) == type(v)

def test_parse_subscription_list():
    """
    Test parsing of a subscription value response, where the response value is a list
    """
    # Create test publish token
    TEST_SUBSCRIBE_TYPE = "TestType"
    TEST_CHANNEL_ID = "ALL"
    TEST_BLOCK_ID = "LongName_Block_With_Many_Underscores"

    # f"S_{subscribe_type}_{channel_id}_{self._block_id}")
    TEST_PUBLISH_TOKEN = f"S_{TEST_SUBSCRIBE_TYPE}_{TEST_CHANNEL_ID}_{TEST_BLOCK_ID}"

    for v1 in [True, False, 9.0, -9.0, 1, -1, "stringa"]:
        for v2 in [True, False, 9.0, -9.0, 1, -1, "stringb"]:

            v1_str = str(v1).lower().strip()
            v2_str = str(v2).lower().strip()

            r = TTPResponse(f'! "publishToken":"{TEST_PUBLISH_TOKEN}" "value":[{v1_str} {v2_str}]')
            assert r.type == TTPResponseType.SUBSCRIPTION
            assert r.publish_token == TEST_PUBLISH_TOKEN
            assert r.subscription_type == TEST_SUBSCRIBE_TYPE
            assert r.subscription_channel_id == TEST_CHANNEL_ID
            assert r.subscription_block_id == TEST_BLOCK_ID
            assert type(r.value) == list
            assert len(r.value) == 2

            assert type(r.value[0]) == type(v1)
            assert type(r.value[1]) == type(v2)
            assert r.value[0] == v1
            assert r.value[1] == v2

def test_parse_subscription_singlevalue():
    """
    Test parsing of a subscription value response, where the response value is a singular value
    """
    # Create test publish token
    TEST_SUBSCRIBE_TYPE = "TestType"
    TEST_CHANNEL_ID = "ALL"
    TEST_BLOCK_ID = "BlockName With Spaces"

    # f"S_{subscribe_type}_{channel_id}_{self._block_id}")
    TEST_PUBLISH_TOKEN = f"S_{TEST_SUBSCRIBE_TYPE}_{TEST_CHANNEL_ID}_{TEST_BLOCK_ID}"

    def send_and_check(val_str, expected_val_typed):
        r = TTPResponse(f'! "publishToken":"{TEST_PUBLISH_TOKEN}" "value":{val_str}')
        assert r.type == TTPResponseType.SUBSCRIPTION
        assert r.publish_token == TEST_PUBLISH_TOKEN
        assert r.subscription_type == TEST_SUBSCRIBE_TYPE
        assert r.subscription_channel_id == TEST_CHANNEL_ID
        assert r.subscription_block_id == TEST_BLOCK_ID
        assert type(r.value) == type(expected_val_typed)
        assert r.value == expected_val_typed

    # Check different value types
    for v in [True, False, 0, 1, 2.0, -2.0, "blah"]:

        # We string-encode for sending
        string_encoded = str(v).lower().strip()

        # Now we check the value in different ways: returned directly,
        # or enclosed in double quotes
        send_and_check(string_encoded, v)
        send_and_check(f'"{string_encoded}"', v)

def test_value_formatter():
    """
    Test the value formatter. This is a bit hidden, so we
    bring it out before we do so...

    Note: this is a standalone method and the self paramter
          isn't used, so we just pass None into it
    """
    value_format = lambda v: TTPResponse._TTPResponse__value_format(None, v)

    # Boolean
    for v in ["true", "yes", "on"]:
        assert value_format(v) == True
    for v in ["false", "no", "off"]:
        assert value_format(v) == False

    # Integer
    for i in range(-10, 10):
        assert value_format(f"{i}") == i
        assert type(value_format(f"{i}")) == int

    # Floating-point
    for i in range(-10, 10):
        fv = float(i * 0.1)
        assert value_format(f"{fv}") == fv
        assert type(value_format(f"{fv}")) == float

    # String
    assert value_format("string") == "string"
    assert type(value_format("string")) == str

def test_parse_active_fault_list():
    """
    Test parsing of active fault list
    """
    RESPONSE = r'+OK "value":[{"id":INDICATOR_NONE_IN_DEVICE "name":"No fault in device" "faults":[] "serialNumber":"00000000"}]'
    r = TTPResponse(RESPONSE)
    assert r.type == TTPResponseType.CMD_OK_VALUE

    assert type(r.value) == list
    assert len(r.value) == 1
    
    rv = r.value[0]
    assert rv["id"] == "INDICATOR_NONE_IN_DEVICE"
    assert rv["name"] == "No fault in device"
    assert rv["faults"] == []
    assert rv["serialNumber"] == 0

def test_parse_network_status():
    """
    Test parsing of a network status response
    """
    RESPONSE = r'+OK "value":{"schemaVersion":2 "hostname":"TestDSP" "defaultGatewayStatus":"0.0.0.0" "networkInterfaceStatusWithName":[{"interfaceId":"control" "networkInterfaceStatus":{"macAddress":"00:00:00:00:00:00" "linkStatus":LINK_1_GB "addressSource":STATIC "ip":"192.168.1.2" "netmask":"255.255.255.0" "dhcpLeaseObtainedDate":"N\/A" "dhcpLeaseExpiresDate":"N\/A" "gateway":"192.168.1.1"}}] "dnsStatus":{"primaryDNSServer":"192.168.1.3" "secondaryDNSServer":"0.0.0.0" "domainName":""} "mDNSEnabled":true "telnetDisabled":true "sshDisabled":false "networkPortMode":PORT_MODE_SEPARATE "rstpEnabled":false "httpsEnabled":false "igmpEnabled":false "switchPortMode":SWITCH_PORT_MODE_CONTROL_AND_MEDIA}'

    # Get basic response
    r = TTPResponse(RESPONSE)
    assert r.type == TTPResponseType.CMD_OK_VALUE

    assert r.value["schemaVersion"] == 2
    assert r.value["hostname"] == "TestDSP"
    assert r.value["defaultGatewayStatus"] == "0.0.0.0"

    assert type(r.value["dnsStatus"]) == dict
    assert r.value["dnsStatus"]["domainName"] == ""
    assert r.value["dnsStatus"]["primaryDNSServer"] == "192.168.1.3"
    assert r.value["dnsStatus"]["secondaryDNSServer"] == "0.0.0.0"

    assert r.value["networkPortMode"] == "PORT_MODE_SEPARATE"
    assert r.value["rstpEnabled"] == False
    assert r.value["sshDisabled"] == False
    assert r.value["telnetDisabled"] == True
    assert r.value["switchPortMode"] == "SWITCH_PORT_MODE_CONTROL_AND_MEDIA"

    assert type(r.value["networkInterfaceStatusWithName"]) == list
    assert len(r.value["networkInterfaceStatusWithName"]) == 1

    assert r.value["networkInterfaceStatusWithName"][0]["interfaceId"] == "control"
    assert type(r.value["networkInterfaceStatusWithName"][0]["networkInterfaceStatus"]) == dict

    assert r.value["networkInterfaceStatusWithName"][0]["networkInterfaceStatus"]["addressSource"] == "STATIC"
    assert r.value["networkInterfaceStatusWithName"][0]["networkInterfaceStatus"]["gateway"] == "192.168.1.1"
    assert r.value["networkInterfaceStatusWithName"][0]["networkInterfaceStatus"]["ip"] == "192.168.1.2"
    assert r.value["networkInterfaceStatusWithName"][0]["networkInterfaceStatus"]["netmask"] == "255.255.255.0"
    assert r.value["networkInterfaceStatusWithName"][0]["networkInterfaceStatus"]["macAddress"] == "00:00:00:00:00:00"
    assert r.value["networkInterfaceStatusWithName"][0]["networkInterfaceStatus"]["linkStatus"] == "LINK_1_GB"


#!/usr/bin/env python3
#
# Run this with python3 -i to get interactive debug shell!
#
from pytesira.dsp import DSP
from pytesira.transport.ssh import SSH
import yaml, time, logging, sys

# Logging configuration
debug = False
logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)
try:
    if sys.argv[1] == "debug":
        logging.getLogger().setLevel(logging.DEBUG)
        debug = True
except:
    pass

with open("config.yaml", "r") as f:
    config = yaml.safe_load(f)

device = DSP(block_map = "dsp_test.bmap")
device.connect(backend = SSH(
                        hostname = config["connection"]["host"],
                        username = config["connection"]["username"], 
                        password = config["connection"]["password"],
                        host_key_check = False
                ))
device.save_block_map(output = "dsp_test.bmap")


#!/usr/bin/env python3
#
# PyTesira DSP class (the one thing that people will import)
#

# Base transport and block classes
from pytesira.transport.transport import Transport
from pytesira.block.block import Block
from pytesira.util.ttp_response import TTPResponse, TTPResponseType

# Others
import os
import time
import json
import queue
import logging
import importlib
from threading import Thread, Event

class DSP:

    """
    PyTesira DSP class - the main thing!


    TODO:
        * Graceful handling of SSH connection restarts - should also trigger re-subscription on blocks

    """

    def __init__(self,
        block_map : str | None = None,          # Block map (attributes cache) file location. This allows PyTesira to bypass block query
                                                # on startup, resulting in much faster initialization, but will result in incorrect attributes
                                                # if the system is reprogrammed through the Tesira software. If left blank, or there are
                                                # detectable differences (e.g., different number of DSP blocks), PyTesira will automatically
                                                # query the DSP for attribute updates, which may take some time (especially on complex setups)
        device_refresh_interval : int = 5,      # Device data refresh interval - how often should we poll for things like active alarms?
    ) -> None:

        # PyTesira version
        self.__version = "0.1.0"

        # Ready for operation?
        self.ready = False

        # Get a local logger first
        self.__logger = logging.getLogger(__name__)

        # Exit event for child threads, to coordinate a clean shutdown
        self.__exit = Event()

        # Status flags
        self.__connected = Event()              # are we connected to the DSP? If not, many operations are not possible, and any data 
                                                # we may have cached might be stale. In that case, read/write access to DSP block objects 
                                                # should raise an exception. Additionally, if the transport status has changed (say, a
                                                # re-connection has occured), DSP blocks will need to re-subscribe to TSP data streams.

        # Buffers
        self.__rx_buffer = ""                   # "raw" buffer for TTP (Tesira Text Protocol) data we received from device
        self.__rx_cmd_mailbox = None            # "command response" mailbox - here we only handle one command at a time!

        # Block map file
        self.__block_map = block_map

        # Device data refresh interval
        assert 1 <= int(device_refresh_interval), "invalid data refresh interval, must be >= 1"
        self.__device_data_refresh_interval = int(device_refresh_interval)

    # =================================================================================================================

    def connect(self,
        backend: Transport,             # Backend connection transport to the DSP itself (e.g., SSH, Telnet)
    ) -> None:
        """
        Connect to the DSP
        """
        assert not self.__connected.is_set(), "Already connected"

        # Start transport channel
        self.__transport = backend
        self.__transport.start(exit_event = self.__exit, connected_flag = self.__connected)

        # Wait for channel to start
        self.__logger.info("waiting for TTP channel")
        while not self.__connected.is_set():
            time.sleep(0.01)

        # Start listen loop
        self.__rx_thread_handle = Thread(target = self.__rx_loop)
        self.__rx_thread_handle.start()

        # Outgoing command queue, this is used so each block can append to it with synchronous
        # DSP command strings it wanted to process, which the main thread (us!) will process
        # in the order in which it was received, to prevent concurrent synchronous access
        self.__command_queue = queue.Queue()

        # Mailbox for synchronous commands executed in the main thread
        self.__sync_cmd_mailbox = None

        # Start thread to process synchronous command requests in local queue
        # after this is started, __sync_command() can and should be used for every
        # TTP command that is to be sent out to the DSP
        self.__sync_cmd_thread_handle = Thread(target = self.__sync_cmd_process_loop)
        self.__sync_cmd_thread_handle.start()

        # Session baseline setup
        self.__sync_command("SESSION set verbose true")
        self.__sync_command("SESSION set detailedResponse false")

        # Basic device statistics
        self.hostname = str(self.__sync_command("DEVICE get hostname").value)
        self.software_version = str(self.__sync_command("DEVICE get version").value)
        self.serial_number = str(self.__sync_command("DEVICE get serialNumber").value)

        # Query DSP aliases/blocks
        self.__dsp_aliases = list(self.__sync_command("SESSION get aliases").value)

        # Information logging for device parameters
        self.__logger.info(f"Connected to '{self.hostname}' (S/N {self.serial_number}; software version {self.software_version}; {len(self.__dsp_aliases)} DSP aliases)")
        
        # Discovered servers (in configuration)
        self.discovered_servers = self.__sync_command("DEVICE get discoveredServers").value

        # Get DSP block map, either from block map file (if loadable and correct for our DSP) or live query
        self.__block_map = self.__getDSPBlockMap()

        # Subscriptions
        self.__subscriptions = {}

        # Then, for each object in the block map, we initialize it
        self.blocks = {}
        for block_id, block in self.__block_map.items():

            # Get block type
            block_type = block["type"]

            # Is this block type even supported?
            try:

                # Yes, we get a module handle for that block type
                block_module = importlib.import_module(f"pytesira.block.{block_type}", "pytesira")
                block_module_version = str(getattr(block_module, f"{block_type}").VERSION)

                # Pre-mapped "attribute map / initialization helper" to pass along to module initialization?
                block_module_init_helper = None
                if "attributes" in block and type(block["attributes"]) == dict and len(block["attributes"]) >= 1:
                    try:
                        assert block["attributes"]["version"] == block_module_version, "block module version mismatch"
                        assert type(block["attributes"]["helper"]) == dict, "invalid block module map type"
                        block_module_init_helper = block["attributes"]["helper"]

                        # Empty helpers shouldn't be loaded
                        if len(block_module_init_helper) <= 0:
                            block_module_init_helper = None

                    except Exception as e:
                        self.__logger.debug(f"block {block_id} ({block_type}) attribute helper cannot be loaded: {e}")

                # Then, we initialize the module
                self.blocks[block_id] = getattr(block_module, f"{block_type}")(
                    block_id = block_id,                            # block ID on Tesira
                    exit_flag = self.__exit,                        # exit flag to stop the block's threads (sync'd with everything else)
                    connected_flag = self.__connected,              # connected flag (module can refuse to allow access if this is not set)
                    command_queue = self.__command_queue,           # command queue (to run synchronous commands and get results)
                    subscriptions = self.__subscriptions,           # subscriptions container (for routing purposes)
                    init_helper = block_module_init_helper,         # initialization helper, to be used however the module wants
                )

                # At this point we should have initialization helper dict, so we update the block map with that
                self.__block_map[block_id]["attributes"] = self.blocks[block_id].export_init_helper()

                # Set up subscriptions for the module
                self.blocks[block_id].subscribe()

            # No... this module doesn't exist (not supported), so we don't load that
            except ModuleNotFoundError:
                self.__logger.debug(f"Unsupported DSP block type '{block_type}': {block_id}")

        # Start device attribute update loop
        self.__device_data_refresh_loop_handle = Thread(target = self.__device_data_refresh_loop)
        self.__device_data_refresh_loop_handle.start()
        
        # Now we're done - DSP object should now be ready to use!
        self.ready = True
        self.__logger.info("DSP ready")

    # =================================================================================================================

    def close(self) -> None:
        """
        Property terminate connection to the DSP
        """

        # No longer ready
        self.ready = False

        # Set exit flag (should terminate everything)
        self.__exit.set()

        # Wait for threads to terminate
        self.__rx_thread_handle.join()
        self.__sync_cmd_thread_handle.join()
        self.__device_data_refresh_loop_handle.join()

        # Done
        return

    # =================================================================================================================

    def __getDSPBlockMap(self) -> dict:
        """
        Get DSP block map, either from the block map file (if one was specified at time of object creation),
        or live query (if none is provided, or there's an error, or the provided block map doesn't line up with
        DSP attributes we've seen so far)

        MUST be called after we know device hostname and has queried for device aliases,
        otherwise this will fail!
        """

        # Return block map
        rtn_map = {}

        # Are we able to load the block map?
        block_map_loaded = False

        # Try to load block map, if one is specified
        if self.__block_map is not None:
            try:
                self.__block_map = os.path.realpath(self.__block_map)
                with open(self.__block_map, "r") as f:
                    bm = json.load(f)

                    assert bm["hostname"] == self.hostname, "hostname mismatch"
                    assert sorted(bm["aliases"]) == sorted(self.__dsp_aliases), "aliases mismatch"
                    assert bm["pytesira_version"] == self.__version, "PyTesira version mismatch"

                    rtn_map = bm["blocks"]
                    block_map_loaded = True
                    self.__logger.info(f"loaded DSP block map from {self.__block_map}")

            except Exception as e:
                self.__logger.warning(f"could not load DSP block map: {e}")
        
        # If block map isn't yet loaded, there's probably an error in the process somewhere,
        # so we proceed with live query to get latest info
        if not block_map_loaded:
            self.__logger.info("starting DSP block map query")

            # First query for DSP blocks
            for i, block_id in enumerate(self.__dsp_aliases):

                # No need to process device handle
                if block_id == "device":
                    continue

                # Intentionally make an invalid query such that we get the handler response
                # which will specify what type of block this is
                block_type_query_response = self.__sync_command(f'{block_id} get BLOCKTYPE')
                assert block_type_query_response.type == TTPResponseType.CMD_ERROR, "block type query: invalid response type"

                if "::Attributes" not in block_type_query_response.value:
                    # DSP block with no attribute handle... shouldn't happen, 
                    # but if it does we just skip it
                    continue

                # Figure out block interface type, take stuff after last space
                resp = block_type_query_response.value.split(" ")[-1].strip()
                block_type = str(resp).replace("Interface::Attributes", "").strip()
                rtn_map[block_id] = {
                    "type" : str(block_type),       # block type, very important
                    "attributes" : None             # no attributes yet (this will trigger attribute query on supported blocks)
                }
                self.__logger.debug(f"(DSP block discovery: {i + 1}/{len(self.__dsp_aliases)}) {block_id} -> {block_type}")

            self.__logger.info(f"found {len(rtn_map)} DSP blocks")

        # Now we return the block map, and we're done
        return rtn_map

    # =================================================================================================================

    def save_block_map(self, output : str) -> None:
        """
        Save active DSP block map to a file
        """
        output = str(output).strip()
        assert output != "", "Output must be specified"

        # Enforce file extension to prevent confusion and inadvertent inclusion in repository
        if not output.endswith(".bmap"):
            output = f"{output}.bmap"

        output = os.path.realpath(str(output))
        assert self.__block_map, "No active DSP block map"
        with open(output, "w") as f:
            json.dump({
                "hostname" : self.hostname,
                "aliases" : sorted(self.__dsp_aliases),
                "blocks" : self.__block_map,
                "pytesira_version" : self.__version
            }, f, indent = 4)
        self.__logger.info(f"DSP block map saved: {output}")
            
    # =================================================================================================================

    def __sync_cmd_process_loop(self) -> None:
        """
        Synchronous command worker loop. This processes the synchronous command queue,
        which the block objects use to run one-off commands to the DSP and get results.
        """
        self.__logger.info("starting synchronous command processor loop")
        while not self.__exit.is_set():

            # Grab command item from the queue (and expand the tuple)
            try:
                handle, command = self.__command_queue.get(timeout = 0.5)
            except queue.Empty:
                # We didn't get anything, let's just move on to another iteration
                # (by doing so, this evaluates the exit flag at least once every
                #  0.5 seconds, allowing us to kill the thread if needed)
                continue

            # Run the command and return to the correct handle's callback
            # this should be the ONLY place we call transport_send_and_wait!
            response = self.__transport_send_and_wait(command)

            if handle == self:
                # Hey, this is called by us directly, we don't need
                # to callback for this, just write data to local synchronous
                # command mailbox and we'll be done:
                self.__sync_cmd_mailbox = response
                self.__logger.debug(f"sync command response delivery: main loop sync_cmd_mailbox")
            else:
                # This is called from elsewhere, so we need to invoke
                # the corresponding block's callback:
                handle._sync_command_callback(data = response)
                self.__logger.debug(f"sync command response delivery: callback to {handle}")

            # Notify queue of a completed task
            self.__command_queue.task_done()

        # If we're here, we're exiting
        self.__logger.debug("synchronous command processor loop terminated")
        return

    # =================================================================================================================

    def __sync_command(self, command : str, timeout : float = 3.0) -> TTPResponse:
        """
        Command send-and-wait function that uses the main command queue instead of talking
        to the transport channel directly. This should be used after __sync_cmd_process_loop() has
        started running, as it guarantees that only one synchronous command is ever executed at
        a time, preventing conflicts
        """

        # Clean out mailbox
        self.__sync_cmd_mailbox = None

        # Queue command
        self.__command_queue.put((self, command))
        cmd_queued = time.perf_counter()

        # Now we wait until we either get data or timeout occurs, whichever comes first
        while time.perf_counter() - cmd_queued < timeout:

            # Let other tasks run
            time.sleep(0.01)

            # Did we get something?
            if self.__sync_cmd_mailbox:
                cmd_response = self.__sync_cmd_mailbox
                self.__sync_cmd_mailbox = None
                return cmd_response

        # If we're here, we got... nothing
        self.__sync_cmd_mailbox = None
        raise Exception(f"Command timeout: {command}")

    # =================================================================================================================

    def __rx_loop(self) -> None:
        """
        Receive loop, to be run as a thread. Constantly monitors and receives stuff
        from the backend TTP connection.

        Also handles subscription routing (to the right DSP block handler) and
        mailboxing for synchronous (non-subscription) commands

        """
        self.__logger.info("starting TTP receiver loop")
        while not self.__exit.is_set():

            # Let backend I/O run
            time.sleep(0.00001)

            rx = self.__transport_rx()
            if rx:
                # Get something, let's try to parse that
                decoded = TTPResponse(rx)

                # Subscription?
                if decoded.type == TTPResponseType.SUBSCRIPTION:
                    self.__logger.debug(f"rx (sub): {decoded}")

                    # Where did this come from?
                    subscriber = self.__subscriptions[decoded.publish_token]
                    if not subscriber:
                        self.__logger.error(f"subscription callback for invalid subscriber: {decoded.publish_token}")
                    else:
                        # OK, we have a subscriber, let's call their callback so the block code
                        # can figure out what to do with that data
                        sub_handle, _, _, _ = subscriber
                        sub_handle.subscription_callback(decoded)

                else:
                    self.__logger.debug(f"rx (res): {decoded}")
                    self.__rx_cmd_mailbox = decoded                    

        # If we're here, we're exiting
        self.__logger.debug("rx loop terminated")
        return

    # =================================================================================================================

    def __device_data_refresh_loop(self) -> None:
        """
        Device data refresh loop. This constantly queries the device for attributes that cannot
        be subscribed to, but nevertheless would be of interest to PyTesira users (e.g., active alarms)

        We simply poll the device (with a configurable interval) for this...
        """
        self.__logger.info("starting device data refresh loop")
        while not self.__exit.is_set():

            # Query active faults
            self.faults = self.__sync_command("DEVICE get activeFaultList").value

            # Query network status
            self.network = self.__sync_command("DEVICE get networkStatus").value

            # Throttle query to the interval we're configured for
            # (typicaly 5 seconds, but can be configured to be less or more as needed)
            time.sleep(self.__device_data_refresh_interval)

        # Exit
        self.__logger.debug("device data refresh loop terminated")
        return

    # =================================================================================================================

    def __transport_send_and_wait(self, command : str, timeout : float = 3) -> TTPResponse:
        """
        Send TSP command to device and wait for response (with a timeout limit)
        Note: only ONE command should be sent at a time, rapid-fire sends may
              cause the mailbox to get and return wrong data!
        """

        # Send command and mark time sent
        self.__transport.send(command)
        cmd_sent = time.perf_counter()

        # Now we wait until we either get data or timeout occurs, whichever comes first
        while time.perf_counter() - cmd_sent < timeout:

            # Let other tasks run
            time.sleep(0.01)

            # Did we get something?
            if self.__rx_cmd_mailbox:
                cmd_response = self.__rx_cmd_mailbox
                self.__rx_cmd_mailbox = None
                return cmd_response

        # If we're here, we got... nothing
        self.__rx_cmd_mailbox = None
        raise Exception(f"Command timeout: {command}")

    # =================================================================================================================

    def __transport_rx(self) -> str | None:
        """
        Transport receive handler. Connects to the backend transport stream, then tries
        to extract one newline-delimited line from that stream buffer. Call this multiple
        times to get everything

        Note: before calling this function (repeatedly if needed), be sure to do
        a time.sleep(0.00001), so the backend I/O threads (e.g., Paramiko) gets a
        chance to run and return data!
        """

        # Fill buffer from stream
        while self.__transport.recv_ready:
            self.__rx_buffer += str(self.__transport.recv(self.__transport.read_buffer_size))

        # If there us a newline, response has ended. We go through lines until we get something 
        # with either "+OK", "-ERR", or "!" prefixes, those are valid lines. We don't return
        # invalid lines otherwise.
        #
        # Note: we use a while loop here, so this function will run until it either finds a
        # valid line, or it runs out of newlines, whichever comes first
        while "\n" in self.__rx_buffer:
            newline_position = self.__rx_buffer.find("\n")
            line = str(self.__rx_buffer[:newline_position]).strip()
            self.__rx_buffer = self.__rx_buffer[newline_position + 1:]

            if line != "" and (line.startswith("+OK") or line.startswith("-ERR") or line.startswith("!")):
                # Valid line
                return line

        # Out of newlines, so just return nothing
        return None

    # =================================================================================================================

    def reboot(self) -> None:
        """
        Send reboot (restart) command to device
        """
        assert self.ready, "DSP not ready"
        self.__sync_command("DEVICE reboot")
        return